"""This module defines base classes used to implement script management for cyrano.

$Id: bases.py 486 2008-10-17 20:03:36Z jmiller $
"""

class PseudoIterator(object):
    """A PseudoIterator provides iterator-like operators for a sequence including:
         next(), prev(), first(), last(), current(), [], len().
    """
    def __init__(self, seq, i=0):
        self.i = i
        self._seq = seq

    def first(self):
        if len(self._seq) > 0:
            self.i = 0
            return self._seq[0]
        else:
            raise IndexError("iterated sequence is empty.")

    def last(self):
        if len(self._seq) > 0:
            self.i = len(self._seq) - 1
            return self._seq[self.i]
        else:
            raise IndexError("iterated sequence is empty.")

    def current(self):
        if 0 <= self.i < len(self._seq):
            return self._seq[self.i]
        else:
            raise IndexError("current index is invalid")

    def next(self):
        if self.i < len(self._seq)-1:
            self.i += 1
            x = self._seq[self.i]
            return x
        else:
            raise StopIteration

    def prev(self):
        if self.i > 0:
            self.i -= 1
            x = self._seq[self.i]
            return x
        else:
            raise StopIteration

    def __len__(self):
        return len(self._seq)

    def __getitem__(self, i):
        return self._seq[i]

class Exposition(object):
    """An exposition consists of a block of text which describes a sequence
    of executable statements.   Textually,  the descriptive text precedes the
    statements it explains.  Together,  they demonstrate one feature or concept.
    """
    def __init__(self, description, examples):
        self.descr = description
        if examples == []:
            examples = [ Example(code="\n",want="") ]
        self.examples = PseudoIterator(examples)

    def format_example(self, i):
        e = self.examples[i]
        code = e.source
        if code.strip():
            lines = [l for l in code.split("\n") if l]
            code = "  >>> " + "\n  ... ".join(lines) + "\n"
            lines = [l for l in e.want.split("\n")]
            if len(lines) > 1:
                results = "  " + "\n  ".join(lines) + "\n"
            elif lines[0] == "":
                results = ""
            else:
                results = "  " + lines[0]
            return code + results
        else:
            return "\n"

    def __repr__(self):
        return "Exposition(description=%s, examples=%s)" % \
               (self.descr, self.examples._seq)

    def __str__(self):
        """Format an Exposition as a string,  first showing the description,
        and then showing the lines of code.  Place the *-cursor at the next
        example.   Skip over blank lines with the cursor.
        """
        s = ""
        descr = self.descr or ""
        for l in descr.split("\n"):
            s += "  "
            if l.startswith("#"):
                s += l[1:] + "\n"
            else:
                s += l + "\n"

        # s += "\n"

        nocursor = True
        for i in range(len(self.examples)):
            estr = self.format_example(i)
            if i < self.examples.i:
                s += estr
            elif nocursor and estr.strip():
                s += "*" + estr[1:-1] + " \n"
                nocursor = False
            else:
                s += estr
        return s + "\n"

class Example(object):
    """A dumbed down class based on the doctest Example class which represents both
    doctest and linear script oriented demos using the same interface.  It has an attributes
    which contain a string of executable Python code as well as a string of the expected
    results.
    """
    def __init__(self, code, want=""):
        self.source = code
        self.want = want

    def __repr__(self):
        return "Example(code=%s, want=%s)" % (repr(self.source), repr(self.want))

class ScriptManager(object):
    """A script manager is associated with a named file which scripts a demonstration
    of a Python package.   The script manager contains an execution context as well as
    a sequence of Exposition objects which can be iterated through.
    """
    def __init__(self, name="", context={},  expositions=[]):
        self.name = name
        self.context = context
        self.expositions = PseudoIterator(expositions)
