#! /usr/bin/env python

"""
Cyrano is a tool for presenting doctests or scripts as demonstrations
or oral presentations.  The basic idea is that you script a demo in
doctest format, and then Cyrano will display your doctest and step
through it as you discuss it for an audience.  In addition, Cyrano
provides access in its interactive shell to the doctest context, so
you can jump in and insert additional code or try things at any point.

Although doctest does provide a format which supports easy verification
of the examples you're going to present,  and cues the audience as to
what output to expect,  simple scripts are also supported.

The structure of Cyrano demos departs slightly from doctest by
grouping a single description with multiple executable statements into
what I will call an 'Exposition'.  A Cyrano demo is a sequence of
Expositions.  The idea of an Exposition is that it contains enough
statements to make one point or demonstrate one thing, generally a
bunch of setup statements followed by the coup d'gras.  So an
Exposition groups multiple doctest Examples into a more useful entity.

On the first step of an Exposition, the description and all the
statements are displayed.  On each additional step, one statement is
executed by 'ghost typing' into the interactive shell.  Pressing the
fast forward button executes all statements in the Exposition.
Pressing the run button executes all of the Expositions.  Pressing the
stop button eventually halts a fast-forward or run...  but it does not
interrupt.

$Id: cyrano.py 1038 2009-08-07 19:32:56Z jmiller $
"""
import os, sys, doctest, pdb, profile as cProfile, time, copy, os.path, re
import Pmw, Tkinter as Tk, tkFont, FileDialog, tkMessageBox
import tkipshell as tkip

from linear_manager import script_manager_from_linear_script
from doctest_manager import script_manager_from_doctest

# =====================================================================================
from config import *
try:
    RC = os.environ["HOME"]+"/"+".cyranorc"
    rccontext = {}
    execfile(RC, rccontext)
except:
    print >>sys.stderr, "Error in RC file:", repr(sys.exc_info()[1]),"... using cyrano defaults."
    print >>sys.stderr, "To customize, copy site-packages/cyrano/config.py to ${HOME}/.cyranorc"
else:
    print "found rc:", RC
    globals().update(rccontext)

# =====================================================================================
labelFont = Pmw.logicalfont('Helvetica', sizeIncr=4, style="bold", weight="bold")
# buttonFont = Pmw.logicalfont('', weight="bold")

# =====================================================================================

class WithdrawableFrame(Tk.Frame):
    def withdraw(self):
        pass
    def wm_title(self, n):
        pass
    def deiconify(self):
        pass
    def minsize(self, x, y):
        pass
    def quit(self):
        pass
    def destroy(self):
        pass
    def configure(self, *args, **keys):
        pass
    def update(self):
        pass


class Cyrano(object):
    def __init__(self, root, layout="vertical"):
        self._root = root
        self._context = {}
        self._stop = False
        self._locked = 0
        self._page_pending = False
        self._exposition = None
        self._init_gui(root, layout)

    def _init_gui(self, root, layout):

        self._menubar = Pmw.MainMenuBar(root)
        root.configure(menu=self._menubar)

        if EMBED_PYLAB:
            self._demo_plot = Pmw.PanedWidget(
                parent=self._root, orient="horizontal")
            self._demo_root = self._demo_plot.add("demo")
            self._plot = WithdrawableFrame(self._demo_plot.add("plot"))
        else:
            self._demo_root = root
            self._plot = None
            
        self._demo = Pmw.PanedWidget(
            parent=self._demo_root, orient="vertical", handlesize=0, separatorthickness=0)
        self._demo.pack(expand=1, fill="both", side="left")

        if EMBED_PYLAB:
            import embed_pylab
            self._pylab_figure = embed_pylab.figure_widget(master=self._plot)
            self._pylab_figure._tkcanvas.pack(
                expand=1, fill="both", side="left")
            self._plot.pack(expand=1, fill="both", side="left")

        self._demo.add("panes") # , min=200,size=200)
        self._demo.add("buttons", min=32,max=32,size=32)
        self._button_bar = self._demo.pane("buttons")
        self._panes = Pmw.PanedWidget(
            parent=self._demo.pane("panes"), orient=layout)
        self._panes.add("description") #, min=100, size=100)
        self._panes.add("ipython") #, min=100, size=100)

        self._menubar.addmenu("File", None)
        self._menubar.addmenuitem(
            "File", "command", label="Open", command=self.open)
        self._menubar.addmenuitem(
            "File", "command", label="Quit", command=self.quit)

        self._menubar.addmenu("Options", None)

        self._description = Pmw.ScrolledText(
            self._panes.pane("description"),
            text_width=DESCRIPTION_WIDTH,
            text_height=DESCRIPTION_HEIGHT,
            text_background=DESCRIPTION_BACKGROUND,
            text_foreground=DESCRIPTION_FOREGROUND,
            text_state="disabled",
            text_wrap="none")
        text = self._description.component("text")
        text.tag_config("DESCR_TEXT",
            foreground=DESCRIPTION_TEXT,
            background=DESCRIPTION_TEXT_BACKGROUND,
            font=(DESCRIPTION_TEXT_FAMILY,
                  DESCRIPTION_TEXT_SIZE,
                  DESCRIPTION_TEXT_STYLE),
            relief=DESCRIPTION_TEXT_RELIEF)
        text.tag_config("DESCR_CODE",
            foreground=DESCRIPTION_CODE,
            background=DESCRIPTION_CODE_BACKGROUND,
            font=(DESCRIPTION_CODE_FAMILY,
                  DESCRIPTION_CODE_SIZE,
                  DESCRIPTION_CODE_STYLE),
            relief=DESCRIPTION_CODE_RELIEF)
        text.tag_config("READY",
            foreground=DESCRIPTION_READYLINE,
            background=DESCRIPTION_READYLINE_BACKGROUND,
            relief=DESCRIPTION_READYLINE_RELIEF,
            borderwidth=DESCRIPTION_READYLINE_BORDERWIDTH)

        self._tkip = tkip.IPythonView(
            self._panes.pane("ipython"),
            user_ns=self._context, user_global_ns=self._context)
        self._tkip.configure(
            text_width=IPYTHON_WIDTH,
            text_height=IPYTHON_HEIGHT,
            text_wrap="none")

        self._initial_context = copy.copy(self._context)

        self._tkip.component("text").configure(
            foreground=IPYTHON_FOREGROUND, background=IPYTHON_BACKGROUND,
            insertbackground=IPYTHON_CURSOR_COLOR,
            insertwidth=IPYTHON_CURSOR_WIDTH)

        self._buttons     = Tk.Frame(borderwidth=2, master=self._button_bar)
        self._button_grid_col = 1

        self._step_button      = self._add_button(
            "step", self.step, master=self._buttons)
        self._cont_button      = self._add_button(
            "cont", self.cont, master=self._buttons)
        self._run_button       = self._add_button(
            "run",  self.run, master=self._buttons)
        self._stop_button      = self._add_button(
            "stop", self.stop, master=self._buttons)

        self._add_spacer(25)

        self.clear_button = self._add_button(
            "clear", self.clear, master=self._buttons)
        self.reset_button = self._add_button(
            "reset", self.reset, master=self._buttons)

        self._add_spacer(25)

        self._page_up_button   = self._add_button(
            "<<", self.page_up, master=self._buttons)
        self._back_button      = self._add_button(
            "<", self.back, master=self._buttons)
        self._skip_button      = self._add_button(
            ">", self.skip, master=self._buttons)
        self._page_down_button = self._add_button(
            ">>", self.page_down, master=self._buttons)

        self._font_family_menu = self._add_cascade(
            "Options", "Font family", choices=tkFont.families(),
            default=DEFAULT_FONT_FAMILY, command=self._update_font)
        self._font_size_menu = self._add_cascade(
            "Options", "Font size", choices=[str(x) for x in FONT_SIZES],
            default=str(DEFAULT_FONT_SIZE), command=self._update_font)
        self._font_style_menu = self._add_cascade(
            "Options", "Font style", choices=["normal", "bold", "italic", "bold italic"],
            default=DEFAULT_FONT_STYLE, command=self._update_font)

        self._window_wrap_var = Tk.BooleanVar()
        self._window_wrap_var.set(WRAP_LINES)
        self._window_wrap_checkbox = self._menubar.addmenuitem(
            "Options", "checkbutton", label="wrap lines",
            variable=self._window_wrap_var,
            command=self._set_window_wrap)

        self._panes.pack(expand=1, fill="both")
        self._description.pack(expand=1, fill="both", side="top")
        self._tkip.pack(expand=1, fill="both", side="top")

        self._button_bar.pack(expand=0, fill="x", side="top")

        self._buttons.pack(expand=0, fill="x", side="left")
        
        self._panes.updatelayout()
        self._demo.updatelayout()

        if EMBED_PYLAB:
            self._demo_plot.pack(expand=1, fill="both")
            self._demo_plot.updatelayout()

        self._update_font()
        self._set_window_wrap()
        self._tkip.clear()

    def _set_window_wrap(self):
        state = {False:"none", True:"word"}[self._window_wrap_var.get()]
        self._description.component("text").configure(wrap=state)
        self._tkip.component("text").configure(wrap=state)

    def _add_spacer(self, space):
        self._buttons.grid_columnconfigure(
            self._button_grid_col, minsize=space)
        self._button_grid_col += 1

    def _add_button(self, text, command, master=None, pack="left", **kw):
        b = Tk.Button(text=text, master=master, command=command, **kw)
        col = self._button_grid_col
        self._button_grid_col += 1
        b.grid(row=0, column=col, padx=1, sticky="ew")
        # b.pack(side=pack)
        return b

    def _add_cascade(self, menu, cascademenu, choices, default=None,
                     balloon=None, kind="radiobutton", command=None):
        self._menubar.addcascademenu(menu, cascademenu)
        var = Tk.StringVar()
        var.set(default)
        for c in choices:
            try:
                self._menubar.addmenuitem(
                    cascademenu, kind, balloon, label=str(c),
                    variable=var, value=c, command=command)
            except UnicodeEncodeError:  # Need scheme for unicode family names
                pass
        return var

    
    def _font_fields(self, font):
        print >>sys.stderr, "font_size:", repr(font)
        fields = font.split(" ")
        fields2 = []
        E = enumerate(fields)
        for i, f in E:
            if f.startswith("{"):
                f = f[1:]
                while 1:
                    i, g = E.next()
                    f += " " + g
                    if g.endswith("}"):
                        f = f[:-1]
                        break
            fields2.append(f)
        fields = fields2
        family = fields[0]
        size = len(fields)>1 and abs(int(fields[1])) or 12
        style = len(fields)>2 and fields[2] or "normal"
        return family, size, style

    def _update_font(self, *args):
        try:
            ip = self._tkip.component("text")
            text = self._description.component("text")
            
            global_family, global_size, global_style = \
                self._font_fields(text.cget("font"))
            
            family = self._font_family_menu.get()
            size = int(self._font_size_menu.get())
            style = self._font_style_menu.get()
            
            # print >>sys.stderr, "font menu:", family, size, style        
            text.configure(font=(family, size, style))
            ip.configure(font=(family, size, style))
            
            # Adjust size relative to old size and global delta
            # Change family and style only if global changed,  else custom.
            code_font = self._font_fields(text.tag_cget("DESCR_CODE", "font"))
            code_font = (family, size-global_size + code_font[1],
                         DESCRIPTION_CODE_STYLE)
            text.tag_config("DESCR_CODE", font=code_font)
            
            text_font = self._font_fields(text.tag_cget("DESCR_TEXT", "font"))
            text_font = (family, size-global_size + text_font[1],
                         DESCRIPTION_TEXT_STYLE)
            text.tag_config("DESCR_TEXT", font=text_font)
        except:
            print >>sys.stderr, "Error updating font:", sys.exc_info()[1]

    def quit(self):
        self._tkip.quit()

    def open(self):
        dialog = FileDialog.LoadFileDialog(self._root, title="Open file")
        filename = dialog.go(pattern="*")
        if filename:
            self.set_file(filename)

    def clear(self):
        "Effectively erases the IPython display."
        self._notify("clear.")
        self._locked = 0
        self._page_pending = False
        self._tkip.clear()

    def reset(self):
        "Completely resets the entire demo, including execution context."
        self._notify("reset.")
        if not tkMessageBox.askokcancel("Reset","Reset demonstration?"):
            return
        self.set_file(self._manager.name)
        self.clear()

    def set_file(self, name):
        kind = detect_demo_kind(name)
        if kind == "script":
            m = script_manager_from_linear_script(name)
        elif kind == "doctest":
            m = script_manager_from_doctest(name)
        else:
            raise ValueError("Unknown demo type.")
        self.set_manager(m)

    def set_manager(self, manager):
        self._root.title(manager.name)
        self._context.clear()
        self._context.update(self._initial_context)
        self._context.update(manager.context)
        self._manager = manager
        self._exposition = self._manager.expositions.first()
        self.set_description(str(self._exposition))

    def string_pound(self, l):
        """returns True if string `l` contains a # """
        ans = re.match(".*'.*#.*'.*", l)
        ans = ans or re.match('.*".*#.*".*', l)
        return ans

    def execute(self, source, delay=0.010):
         # old = self._tkip.IP.autoindent
        self._tkip.IP.autoindent = 0
        lines = [l for l in source.split("\n")]
        if len(lines)==2 and lines[-1] == "":
            lines = lines[:-1]
        for i, l in enumerate(lines):
            # self._notify("executing: " + str(len(l)) + " " + repr(l))
            l = l.rstrip()
            if l.lstrip().startswith("#"): # indented comment
                continue
            if len(l.split("#")) >= 2: # trailing comment
                if not self.string_pound(l):
                    l = l.split("#")[0]
            if i < len(lines)-1: # Use \ to continue multilines
                l += " \\"
            # l += "\n"
            self.ghost_type(l, delay)
            self._tkip.processLine()
        self._tkip.IP.autoindent = 1

    def set_description(self, descr):
        text = self._description.component("text")
        where = 1.0
        lines = descr.split("\n")
        ready_line = -1
        
        descr_lines = 0
        for l in lines:
            s = l.strip()
            if  not s.startswith("*") and not s.startswith(">>>"):
                descr_lines += 1
            else:
                break
            
        for no, l in enumerate(lines):
            s = l.strip()
            if s.startswith("*"):
                where = (no-2)*1.0/len(lines)
                lines[no] = " " + l[1:]
                ready_line = no+1
                
        self._description.settext("\n".join(lines))
        text.tag_add("DESCR_TEXT", 
                     1.0, "%d.end" % (descr_lines,))
        text.tag_add("DESCR_CODE", 
                     "%d.0" % (descr_lines+1,), Tk.END)
        if ready_line >= 0 and not self._page_pending:
            text.tag_add("READY", 
                         "%d.5" % (ready_line,), 
                         "%d.end" % (ready_line,))
            text.yview_moveto(where)
        self._root.update_idletasks()

    def delay_but_update(self, seconds, update_tick=0.010):
        self._root.update()
        ticks = int(seconds / update_tick)
        for i in range(ticks):
            time.sleep(update_tick)
            self._root.update()
            if self._stop:
                    return

    def ghost_type(self, source, char_delay):
        """Send a line of text to the shell as if the user typed it."""
        source = source.rstrip()
        self._tkip.cursor_to_prompt()
        if char_delay:
            for i in range(len(source)):
                self._tkip.changeLine(source[:i+1])
                self.delay_but_update(char_delay)
        else:
            self._tkip.changeLine(source)
            self.delay_but_update(0)

    def _notify(self, msg):
        # print >>sys.stderr, msg
        return msg

    def stop(self, *args):
        self._stop = True
        self._notify("stop")

    def _locked_method(self, name):
        self._notify(name)
        if self._locked:
            return
        self._locked += 1
        try:
            exec "self." + name + "()"
        finally:
            self._locked -= 1
        self._root.update_idletasks()

    def step(self):
        self._locked_method("_step")

    def cont(self):
        self._locked_method("_cont_button_core")

    def run(self):
        self._locked_method("_run")

    def skip(self):
        self._locked_method("_skip_button_core")

    def back(self):
        self._locked_method("_back")

    def page_up(self):
        self._locked_method("_page_up")

    def page_down(self):
        self._locked_method("_page_down")

    def _step(self, how="execute", char_delay=STEP_CHAR_DELAY, line_delay=0):
        if self._page_pending:
            self._skip_button_core()
            return "paged."
        if self._exposition is None:
            self.set_description("")
            return self._notify("demo complete.")
        example = self._exposition.examples.current()
        result = self._skip()
        if result != "stepped.":
            self._page_pending = True
        if example.source.strip():
            getattr(self, how)(example.source, char_delay)
            self.delay_but_update(line_delay)
            self.set_description(str(self._exposition or ""))
            return result
        else:
            self._notify("skipping blank line.")
            return self._step(char_delay=char_delay)

    def _cont(self, line_delay=CONT_LINE_DELAY):
        self._stop = False
        status = self._step(char_delay=0, line_delay=line_delay)
        while status == "stepped." and not self._stop:
            status = self._step(char_delay=0, line_delay=line_delay)
        return status

    def _cont_button_core(self):
        if self._page_pending:
            self._page_pending = False
            self._skip_button_core()
        else:
            if self._cont(line_delay=CONT_LINE_DELAY) != "stepped.":
                self._page_pending = True

    def _run(self, *args):
        self._stop = False
        status = self._cont()
        while status != "demo complete." and not self._stop:
            self.delay_but_update(RUN_EXPOSITION_DELAY)
            status = self._cont(line_delay=RUN_LINE_DELAY)

    def _page_up(self):
        try:
            self._exposition = self._manager.expositions.prev()
        except StopIteration:
            self._exposition = self._manager.expositions.first()
        self._exposition.examples.first()
        self._page_pending = False
        self.set_description(str(self._exposition))

    def _page_down(self):
        try:
            self._exposition = self._manager.expositions.next()
        except StopIteration:
            self._exposition = None
        if self._exposition is not None:
            self._exposition.examples.first()
        self._page_pending = False
        self.set_description(str(self._exposition or ""))

    def _back(self):
        if self._exposition is None: # handle "demo complete" case.
            self._page_up()
            self._exposition.examples.last()
        else:
            try:
                self._exposition.examples.prev()
            except StopIteration:
                self._page_up()
                self._exposition.examples.last()
        self._page_pending = False
        self.set_description(str(self._exposition))

    def _skip(self):   # Widely used advance method
        try:
            example = self._exposition.examples.next()
            if not example.source.strip():
                return self._skip()
            return "stepped."
        except StopIteration:
            return "next exposition."

    def _skip_button_core(self):   # skip button core implementation,  lock method
        if self._skip() != "stepped.":
            self._page_down()
        self._page_pending = False
        self.set_description(str(self._exposition or ""))

    def execute_profile(self, source, *args):
        cProfile.runctx(source, globals=self._context, locals=self._context)

    def _profile(self):
        self._step(how="execute_profile")

    def profile(self):
        self._locked_method("_profile")

def detect_demo_kind(name):
    """Returns true if module includes a doctest.
    """
    global OPTLIST
    if ("-s","") in OPTLIST:
        return "script"
    if ("-d","") in OPTLIST:
        return "doctest"
    if name.endswith(".pyc"):
        return "doctest"
    for l in open(name).readlines():
        if l.lstrip().startswith(">>>"):
            print >>sys.stderr, "Doctest detected... running in doctest mode."
            return "doctest"
    else:
        return "script"


def cyrano(name, layout="vertical"):
    import time
    root = Pmw.initialise(root=Tk.Tk())
    c = Cyrano(root, layout)
    c.set_file(name)
    Pmw.setgeometryanddeiconify(root, WINDOW_GEOMETRY)
    return root, c

OPTLIST = []
ARGS = []

def _main():
    import getopt

    global OPTLIST, ARGS

    OPTLIST, ARGS = getopt.getopt(sys.argv[1:], "dshv")

    if len(ARGS) != 1:
        print >>sys.stderr, "usage:", sys.argv[0], "[-dshv] <demo_file>"
        print >>sys.stderr, "      ", "-d  doctest based demo"
        print >>sys.stderr, "      ", "-s  script based demo"
        print >>sys.stderr, "      ", "-h  horizontal arrangement"
        print >>sys.stderr, "      ", "-v  vertical arrangement"
        sys.exit(1)

    if ("-h","") in OPTLIST:
        layout = "horizontal"
    elif ("-v","") in OPTLIST:
        layout = "vertical"
    else:
        layout = WINDOW_ARRANGEMENT

    root, c = cyrano(ARGS[0], layout=layout)

    return root, c

def main():
    root, c = _main()
    root.mainloop()

if __name__ == "__main__":
    main()


